'
' utf8.bi
'
' UTF-8 library for FreeBASIC, helper and conversion functions
'
' #note: currently only fully supporting the 16 bit range - "plane 0"/"Basic Multilingual Plane"
'        higher planes might not lead to the intended results
'
' UTF8(...) - turns a list of (up to 8) integer values (Unicode) into a string of UTF8-Expressions
'                    returns empty string on error (invalid codes)
'                    note: supports nullbytes but this could cause trouble because those are often
'                    used to mark the end of strings
'
' UTF8FILL(INTEGER multiplier, ULONG Unicode) AS STRING - returns a string filled with
'                    the given Unicode char UTF-8 encoded repeated 'multiplier' times
'                    almost like STRING() - also works with ASCII codes - below 128.
'
' UTF8LEN(STRING string) - returns length of a UTF-8 string in characters, rather than bytes
'                    tries to count only visible characters/text, ignores most, "invisible", control
'                    characters and BOM code
'                    double-width chars should count as 2 chars
'                    not safe, requires 100% valid UTF8 strings
'                    only makes sense to get the length/size of a text line - maybe words,
'                    UTF-8 strings can be handled flawlessly with ASCII functions, as long
'                    as you don't cut the strings haphazardly, i.e. inside UTF-8 sequences
'
' UTF8GETCODE(STRING string[, INTEGER position[, INTEGER offset]]) - returns the Unicode value
'                    of the first character found from given position in 'string', starts
'                    searching from 'position', MID() format => byte position in string starting
'                    with 1
'                    'offset' is an offset in characters to the position in bytes, 1 is the
'                    next character, -1 the previous to 'position', 'position' must point to
'                    an ASCII character or UTF8 startbyte
'                    rather similar to the BASIC function ASC()
'                    #note: I thought it's better to not allow 0 as position value, rather than
'                    counting the value 0 as '1'. It might cause strange errors if you ain't
'                    aware it is 0 resp. if you even don't know you have to use 1 instead of 0
'
' UTF8GETPOS(STRING string, INTEGER position[, INTEGER offset]) - gets the (next valid) char
'                    position in MID() format, returns 0 (is invalid) on error
'                    similar to UTF8GETCODE(), just returning the byte position of the specified
'                    character rather than the characters code
'
' UTF8GETDIFF(STRING string, INTEGER position1[, INTEGER offset1], INTEGER position2[, INTEGER offset2])
'                    - gets the count of characters / length between two given MID()-positions
'                    in a UTF8 string. In principle, it's short for
'                    UTF8LEN(MID(string, position1, position2-position1))
'                    position1 and position2 should point to a valid character/UTF8 start bytes,
'                    position2 is included in the count of characters, double-width chars count
'                    as 2 characters, invisible ones are ignored
'                    you can additionally use offsets in characters, where offset1 is the optional
'                    parameter (when using 4 instead of all 5 parameters)
'                    returns -1 on error, otherwise only the absolute length, if necessary,
'                    you have to determine on your own whether position2 is bigger than position1
'
' UTF8FROMISO8859(STRING string, INTEGER codetable=15) - converts an "ASCII" string into a
'                    UTF-8 string, using translation tables for the ISO/IEC 8859 standard
'                    from ISO 8859-1 to ISO 8859-16
'
' UTF8TOISO8859(STRING source, INTEGER codetable=15) - converts a UTF-8 string into an
'                    ISO 8859 ASCII string; "non-existing" chars are replaced with
'                    non-break space (code 0xA0)
'
' UTF8TOUTF16(STRING source, WSTRING PTR destination, INTEGER size) - converts a UTF-8 string into
'                    a UTF-16 16 bit "wide-char" string, destination must have enough space
'
' UTF8FROMUTF16(WSTRING PTR source) - converts a UTF-16 16 bit "wide-char"
'                    string into a UTF-8 string
'
' UTF8TOUTF32( ) - currently unavailable => inexistant datatype "LSTRING"
'
' UTF8ISVALID(STRING string[, INTEGER start[, INTEGER end, BOOLEAN nonullbytes]]]) - tests a given string whether
'                    all characters are valid UTF8 - returns TRUE if everything is all-right.
'                    This includes plain ASCII, at default ignores nullbytes.
'                    Returns FALSE the first time any illegal code appears.
'                    Position values start+end are in MID() format (first char = 1)
'                    If "nonullbytes" is TRUE, a nullbyte before the string end will also count as invalid.
'                    Prevents memory leaks.
'
' UTF8MID(STRING string, INTEGER position, INTEGER offset[, INTEGER length]) - UTF-8 compatible MID() function
'                    (UTF-8 characters are 2-4 bytes sized, except its ASCII subset => 1 byte)
'                    position is MID()/INSTR()-position/offset in bytes, not characters. First byte is
'                    position "1"
'                    offset lets address a character before or after position if it points to a ASCII or
'                    UTF-8 character, -1 is the character before, +1 is the character following, doesn't
'                    distinguish normal to double width or invisible characters
'                    length is string length in characters, also regarding/cutting double-width chars,
'                    returns rest of string if omitted
'                    if position points to an invalid byte, the returned string starts with the next valid
'                    character, an offset of +1 will point to the same character in this case => ensure,
'                    that position points to the start byte of a UTF-8 code
'
' UTF8LEFT(STRING string, INTEGER toposition[, INTEGER offset]) - UTF-8 version of LEFT() function
'                    toposition is length of substring in bytes, not characters, and points to the last byte
'                    of the substring related to the MID() position value (first byte="1", last byte LEN())
'                    offset is +/- UTF-8 valid chars to this position; toposition must point to last byte of a
'                    UTF-8 sequence or ASCII character
'                    if you want to demand the substring length in characters, use offset instead
'                    and set toposition to 0
'
' UTF8RIGHT(STRING string, INTEGER length) - UTF-8 version of RIGHT() function
'                    returns <length> characters from the end of a string. Length regards double width
'                    characters as 2 chars, all other chars as 1. A character could be bigger than a byte.
'                    UTF-8 characters range from 2 to 4 bytes, only ASCII characters are 1 byte sized.
'
' UTF8DIVIDETOLINES(STRING string, INTEGER linewidth) - divides a UTF8 string into 'lines' with the maximum
'                    given width in chars. Returns one string divided with the divider char '\r' (code 13).
'                    Line width equals the value UTF8LEN() returns, except the invisible line break. Adds hyphens,
'                    supports soft-hyphen, no-break-space and more. If "linewidth" is 20, a word with
'                    e.g. 20 letters will become one line - and will be between 2 '\r's
'                    Only works for monospaced fonts.
'
' UTF8UPPERCASE(STRING string[, [INTEGER start,]INTEGER length]) - tries to capitalise a UTF-8 string, turning
'                    all minuscules into majuscules, length=1 only capitalises the first letter
'                    also working with pure ASCII strings
'
' UTF8LOWERCASE(STRING string[, [INTEGER start,]INTEGER length]) - tries to turn a UTF-8 string into complete lower case,
'                    replacing all majuscules with their minuscules
'                    also working with pure ASCII strings
'
' UNICODECHARWIDTH(UINT32 unicode) / UNICODECHARWIDTH(STRING string[, INTEGER offset]) - returns width of any given
'                    character in "characters". This might be 1 like for most letters, 0 for invisible,
'                    control codes, or 2 for double-width chars, i.e. square sized characters, that have to occupy
'                    2 cells in a monospaced font. The string variant is more efficent compared to determine the
'                    correct Unicode out of a string and retrieving its width; it doesn't parse the UTF8 codes fully
'
' UNICODELOWERCHAR(UINT32 unicode) - returns the Unicode (or ASCII) char that is the lower case form of given char
'                    returns the same code, if already lower case or inapplicable
'
' UNICODEUPPERCHAR(UINT32 unicode) - returns the Unicode (or ASCII) char that is the upper case form of given char
'                    returns the same code, if already upper case or inapplicable
'
' planned/intended: UPPERCASE()/LOWERCASE() (resp. UCASE/LCASE) automatically detecting a UTF8-, UTF16- and
'                    so on string, just in case FreeBASIC would change the STRING type to a variable one with
'                    different encodings

DIM SHARED iso8859tableentry(16, 96) AS ULONG = {_
{&h00A0, &h00A1, &h00A2, &h00A3, &h00A4, &h00A5, &h00A6, &h00A7, &h00A8, &h00A9, &h00AA, &h00AB, &h00AC, &h00AD, &h00AE, &h00AF, _
&h00B0, &h00B1, &h00B2, &h00B3, &h00B4, &h00B5, &h00B6, &h00B7, &h00B8, &h00B9, &h00BA, &h00BB, &h00BC, &h00BD, &h00BE, &h00BF, _
&h00C0, &h00C1, &h00C2, &h00C3, &h00C4, &h00C5, &h00C6, &h00C7, &h00C8, &h00C9, &h00CA, &h00CB, &h00CC, &h00CD, &h00CE, &h00CF, _
&h00D0, &h00D1, &h00D2, &h00D3, &h00D4, &h00D5, &h00D6, &h00D7, &h00D8, &h00D9, &h00DA, &h00DB, &h00DC, &h00DD, &h00DE, &h00DF, _
&h00E0, &h00E1, &h00E2, &h00E3, &h00E4, &h00E5, &h00E6, &h00E7, &h00E8, &h00E9, &h00EA, &h00EB, &h00EC, &h00ED, &h00EE, &h00EF, _
&h00F0, &h00F1, &h00F2, &h00F3, &h00F4, &h00F5, &h00F6, &h00F7, &h00F8, &h00F9, &h00FA, &h00FB, &h00FC, &h00FD, &h00FE, &h00FF},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h2018, &h2019, &h00A3, &h20AC, &h20AF, &h00A6, &h00A7, &h00A8, &h00A9, &h037A, &h00AB, &h00AC, &h00AD, &h0000, &h2015, _
&h00B0, &h00B1, &h00B2, &h00B3, &h0384, &h0385, &h0386, &h00B7, &h0388, &h0389, &h038A, &h00BB, &h038C, &h00BD, &h038E, &h038F, _
&h0390, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h03A0, &h03A1, &h03A2, &h03A3, &h03A4, &h03A5, &h03A6, &h03A7, &h03A8, &h03A9, &h03AA, &h03AB, &h03AC, &h03AD, &h03AE, &h03AF, _
&h03B0, &h03B1, &h03B2, &h03B3, &h03B4, &h03B5, &h03B6, &h03B7, &h03B8, &h03B9, &h03BA, &h03BB, &h03BC, &h03BD, &h03BE, &h03BF, _
&h03C0, &h03C1, &h03C2, &h03C3, &h03C4, &h03C5, &h03C6, &h03C7, &h03C8, &h03C9, &h03CA, &h03CB, &h03CC, &h03CD, &h03CE, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&hFFFE, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000},_
{&h00A0, &h00A1, &h00A2, &h00A3,&h20AC, &h00A5, &h0160, &h00A7, &h0161, &h00A9, &h00AA, &h00AB, &h00AC, &h00AD, &h00AE, &h00AF, _
&h00B0, &h00B1, &h00B2, &h00B3, &h017D, &h00B5, &h00B6, &h00B7, &h017E, &h00B9, &h00BA, &h00BB, &h0152, &h0153, &h0178, &h00BF, _
&h00C0, &h00C1, &h00C2, &h00C3, &h00C4, &h00C5, &h00C6, &h00C7, &h00C8, &h00C9, &h00CA, &h00CB, &h00CC, &h00CD, &h00CE, &h00CF, _
&h00D0, &h00D1, &h00D2, &h00D3, &h00D4, &h00D5, &h00D6, &h00D7, &h00D8, &h00D9, &h00DA, &h00DB, &h00DC, &h00DD, &h00DE, &h00DF, _
&h00E0, &h00E1, &h00E2, &h00E3, &h00E4, &h00E5, &h00E6, &h00E7, &h00E8, &h00E9, &h00EA, &h00EB, &h00EC, &h00ED, &h00EE, &h00EF, _
&h00F0, &h00F1, &h00F2, &h00F3, &h00F4, &h00F5, &h00F6, &h00F7, &h00F8, &h00F9, &h00FA, &h00FB, &h00FC, &h00FD, &h00FE, &h00FF},_
{&h00A0, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
&h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000} _
}

DECLARE FUNCTION UNICODECHARWIDTH OVERLOAD (BYREF character AS ULONG) AS UBYTE
DECLARE FUNCTION UNICODECHARWIDTH OVERLOAD (BYREF source AS STRING, BYREF position AS INTEGER=1) AS UBYTE


FUNCTION UTF8(BYVAL code AS ULONG, BYVAL code1 AS ULONG=&hFFFE, BYVAL code2 AS ULONG=&hFFFE, BYVAL code3 AS ULONG=&hFFFE, BYVAL code4 AS ULONG=&hFFFE, BYVAL code5 AS ULONG=&hFFFE, BYVAL code6 AS ULONG=&hFFFE, BYVAL code7 AS ULONG=&hFFFE) AS STRING
  DIM AS ULONG codeno(8), i
  DIM AS STRING result=""

  codeno(1)=code1 ' crude code for a list of Unicodes, allows to hand up to 8 values at once
  IF code1<>&hFFFE THEN
    codeno(2)=code2
    codeno(3)=code3
    codeno(4)=code4
    codeno(5)=code5
    codeno(6)=code6
    codeno(7)=code7

    i=0
    WHILE code<>&hFFFE ' is invalid Unicode code (mirrored BOM)
      IF code<128 THEN
        ' ASCII range
        result+=CHR(code)
      ELSEIF code<&h0800 THEN
        ' 2 byte code
        result+=CHR(((code SHR 6) OR 192), (code AND 63)+128)
      ELSEIF code<&h10000 THEN
        ' 3 byte code
        result+=CHR(((code SHR 12) OR 224), ((code SHR 6) AND 63)+128, (code AND 63)+128)
      ELSEIF code<&h140000 THEN
        ' 4 byte code
        result+=CHR(((code SHR 18) OR 240), ((code SHR 12) AND 63)+128, ((code SHR 6) AND 63)+128, (code AND 63)+128)
      ELSE
        ' illegal code
        ' result=""
      ENDIF
      i+=1
      code=codeno(i)
    WEND
  ELSE ' faster treatment if just a single value
    IF code<128 THEN
      ' ASCII range
      result+=CHR(code)
    ELSEIF code<&h0800 THEN
      ' 2 byte code
      result+=CHR(((code SHR 6) OR 192), (code AND 63)+128)
    ELSEIF code<&h10000 THEN
      ' 3 byte code
      result+=CHR(((code SHR 12) OR 224), ((code SHR 6) AND 63)+128, (code AND 63)+128)
    ELSEIF code<&h140000 THEN
      ' 4 byte code
      result+=CHR(((code SHR 18) OR 240), ((code SHR 12) AND 63)+128, ((code SHR 6) AND 63)+128, (code AND 63)+128)
    ELSE
      ' illegal code
      ' result=""
    ENDIF
  ENDIF

  RETURN result
END FUNCTION


FUNCTION UTF8FILL(BYVAL multiplier AS INTEGER, BYVAL unicode AS ULONG) AS STRING
  DIM AS STRING result="", ucode

  ucode=UTF8(unicode)

  WHILE multiplier
    result+=ucode
    multiplier-=1
  WEND

  RETURN result
END FUNCTION


FUNCTION UTF8GETCODE(BYVAL sstr AS STRING, BYVAL position AS INTEGER=1, BYVAL offset AS INTEGER=0) AS ULONG
  DIM AS INTEGER i, l
  DIM AS UBYTE c
  DIM AS ULONG result

  IF position<1 THEN RETURN &hFFFE ' invalid nonsense value => error, maybe #todo:

  result=0

  ' find first valid char from 'position' on
  IF position<LEN(sstr) THEN
    ' check forwards
    i=position-1
    l=LEN(sstr)
    WHILE i<l
      c=sstr[i]
      IF c<128 OR c>191 THEN EXIT WHILE ' ignore all UTF-8 "secondary bytes"
      i+=1
    WEND
    IF i=l AND c>127 THEN RETURN &hFFFE ' error: no valid character at position (last byte)
  ELSE ' special case: position=LEN(string), this should address the last character, but
    ' actually just addresses the last byte
    ' locate previous valid char
    i=position-1
    WHILE i
      c=sstr[i]
      IF c<128 OR c>191 THEN EXIT WHILE ' ignore all UTF-8 "secondary bytes"
      i-=1
    WEND
    IF i=l AND c>127 THEN RETURN &hFFFE ' error: no valid character at position (last byte)
  ENDIF

  IF offset<>0 THEN ' check offset (0=none/ignore)
    IF offset>0 THEN
      i+=1
      WHILE i<l
        c=sstr[i]
        IF c<128 OR c>191 THEN
          offset-=1
          IF offset=0 THEN EXIT WHILE
        ENDIF
        i+=1
      WEND
    ELSE
      WHILE i
        i-=1
        c=sstr[i]
        IF c<128 OR c>191 THEN
          offset+=1
          IF offset=0 THEN EXIT WHILE
        ENDIF
      WEND
    ENDIF
  ENDIF

  ' determine & return code
  IF c<128 THEN RETURN c ' ASCII code
  IF c<192 THEN RETURN &hFFFE ' error, c must be a startbyte, propably reached end of string
  ' ELSE: UTF-8 encoded Unicode
  i+=1
  IF (c AND 248)=248 THEN ' 4 byte code
    result = (c AND 7) SHL 6
    l=2
    WHILE l
      c=sstr[i]
      i+=1
      result OR = (c AND 63)
      result SHL = 6
      l-=1
    WEND
    c=sstr[i]
    result OR = (c AND 63)
    RETURN result
  ELSEIF (c AND 240)=224 THEN ' 3 byte code
    result = (c AND 15) SHL 6
    c=sstr[i]
    i+=1
    result OR = (c AND 63)
    result SHL = 6
    c=sstr[i]
    result OR = (c AND 63)
    c=sstr[i]
    result = result AND (c AND 63)
    RETURN result
  ELSEIF (c AND 224)=192 THEN ' 2 byte code
    result = (c AND 31) SHL 6
    c=sstr[i]
    result OR = (c AND 63)
    RETURN result
  ELSE
    ' error, invalid, not UTF-8
  ENDIF

  ' error/fail - return "not a character" code
  RETURN &hFFFE
END FUNCTION


FUNCTION UTF8GETPOS(BYVAL sstr AS STRING, BYVAL position AS INTEGER, BYVAL offset AS INTEGER=0) AS INTEGER
  DIM AS INTEGER i, l
  DIM AS UBYTE c

  IF position<1 THEN RETURN 0 ' invalid nonsense value => error, maybe #todo:

  ' find first valid char from 'position' on
  IF position<LEN(sstr) THEN
    ' check forwards
    i=position-1
    l=LEN(sstr)
    WHILE i<l
      c=sstr[i]
      IF c<128 OR c>191 THEN EXIT WHILE ' ignore all UTF-8 "secondary bytes"
      i+=1
    WEND
    IF i=l AND c>127 THEN RETURN 0 ' error: no valid character at position (last byte)
  ELSE ' special case: position=LEN(string), this should address the last character, but
    ' actually just addresses the last byte
    ' locate previous valid char
    i=position-1
    WHILE i
      c=sstr[i]
      IF c<128 OR c>191 THEN EXIT WHILE ' ignore all UTF-8 "secondary bytes"
      i-=1
    WEND
    IF i=l AND c>127 THEN RETURN 0 ' error: no valid character at position (last byte)
  ENDIF
  IF offset=0 THEN RETURN i+1 ' MID() position of byte

  IF offset>0 THEN
    i+=1
    WHILE i<l
      c=sstr[i]
      IF c<128 OR c>191 THEN
        offset-=1
        IF offset=0 THEN EXIT WHILE
      ENDIF
      i+=1
    WEND
  ELSE
    WHILE i
      i-=1
      c=sstr[i]
      IF c<128 OR c>191 THEN
        offset+=1
        IF offset=0 THEN EXIT WHILE
      ENDIF
    WEND
  ENDIF
  IF offset THEN RETURN 0 ' error (exceeded string limit)
  IF i=l AND c>127 THEN RETURN 0 ' error (reached end of string)

  i+=1 ' MID() position of byte
  RETURN i
END FUNCTION


DECLARE FUNCTION UTF8GETDIFF OVERLOAD (BYVAL source AS STRING, position1 AS INTEGER, position2 AS INTEGER) AS INTEGER
DECLARE FUNCTION UTF8GETDIFF OVERLOAD (BYVAL source AS STRING, position1 AS INTEGER, offset1 AS INTEGER, position2 AS INTEGER, offset2 AS INTEGER) AS INTEGER
FUNCTION UTF8GETDIFF(BYVAL source AS STRING, position1 AS INTEGER, position2 AS INTEGER, offset2 AS INTEGER) AS INTEGER
  RETURN UTF8GETDIFF(source, position1, 0, position2, offset2)
END FUNCTION


FUNCTION UTF8GETDIFF(BYVAL source AS STRING, position1 AS INTEGER, offset1 AS INTEGER, position2 AS INTEGER, offset2 AS INTEGER) AS INTEGER
  DIM AS INTEGER i, l
  DIM AS UBYTE c

  IF source="" OR (position1=position2 AND offset1=offset2) THEN RETURN 0 ' that was easy
  IF position1<1 OR position2<1 THEN RETURN -1 ' corrupt parameters => error
  
  IF offset1 THEN ' (<>0)
    position1=UTF8GETPOS(source,position1,offset1)
  ENDIF
  IF offset2 THEN ' (<>0)
    position2=UTF8GETPOS(source,position2,offset2)
  ENDIF

  IF position1>position2 THEN ' inverse length
    SWAP position1, position2
  ENDIF

  IF position2>LEN(source) THEN
    IF position1>LEN(source) THEN RETURN -1 ' error
    i=LEN(source)
  ELSE
    i=position2
  ENDIF
  ' (still MID() positions/index)

  WHILE i>=position1
    i-=1
    c=source[i]
    IF c<128 THEN ' ASCII
      IF c>31 THEN l+=1 ' visible char / no control code
    ELSEIF c>191 THEN ' UTF8 startbyte
      l+=UNICODECHARWIDTH(source, i+1)
    ENDIF
  WEND

  RETURN l  
END FUNCTION


FUNCTION UTF8GETDIFF(BYVAL source AS STRING, position1 AS INTEGER, position2 AS INTEGER) AS INTEGER
  DIM AS INTEGER i, l
  DIM AS UBYTE c

  IF source="" OR position1=position2 THEN RETURN 0 ' that was easy
  IF position1<1 OR position2<1 THEN RETURN -1 ' corrupt parameters => error
  ' #note: not entirely fail-safe, the positions might be different, but inside
  '        the same UTF-8 encoded character

  IF position1>position2 THEN ' inverse length
    SWAP position1, position2
  ENDIF

  IF position2>LEN(source) THEN
    IF position1>LEN(source) THEN RETURN -1 ' error
    IF position1=LEN(source) THEN RETURN 0
    i=LEN(source)
  ELSE
    i=position2
  ENDIF
  ' (still MID() positions/index) here

  IF i=LEN(source) THEN ' special case: position is LEN => last character
    ' if the last character is UTF-8 encoded, the position points to the
    ' last byte of a sequence instead of the start byte and calculates
    ' a wrong value
    i-=1
    DO
      c=source[i]
      IF c<128 OR c>191 THEN EXIT DO
      i-=1
      IF i<position1 THEN RETURN 0
    LOOP
  ELSE
    i-=1 ' don't count the first character to the difference  
  ENDIF
  
  WHILE i>=position1
    i-=1
    c=source[i]
    IF c<128 THEN ' ASCII
      IF c>31 THEN l+=1 ' visible char / no control code
    ELSEIF c>191 THEN ' UTF8 startbyte
      l+=UNICODECHARWIDTH(source, i+1)
    ENDIF
  WEND

  RETURN l
END FUNCTION


FUNCTION UTF8FROMISO8859(BYVAL source AS string, BYVAL codetable AS integer=15) AS STRING
  ' iso8859tableentry()
  DIM AS STRING result=""
  DIM AS ULONG c
  DIM AS INTEGER i, l

  ' there are 16 "substandards" where #12 was canceled resp. discarded
  IF codetable>1 AND codetable<17 THEN
    codetable-=1 ' 0 equals ISO8859-1
    IF codetable=12 THEN ' also illegal
      RETURN "!ERR_INVALID_CHARSET:"+STR(codetable)
    ENDIF
  ELSE ' illegal value => default ISO8859-1
    RETURN "!ERR_INVALID_CHARSET:"+STR(codetable)
  ENDIF

  i=0
  l=LEN(source)
  WHILE i<l
    c=source[i]
    IF c>127 THEN ' non-ASCII char
      IF c>159 THEN ' ISO-8859 range, use table
        c=iso8859tableentry(codetable, c-160)
        IF c THEN
          result+=UTF8(c)
        ELSE ' translation failed
          result+="?"
        ENDIF
      ELSE ' control chars, 1:1 Unicode
        result+=UTF8(c)
      ENDIF
    ELSE
      result+=CHR(c)
    ENDIF

    i+=1
  WEND

  RETURN result
END FUNCTION


FUNCTION UTF8TOISO8859(BYVAL source AS STRING, BYVAL codetable AS integer=15) AS STRING
  ' #todo:
  RETURN ""
END FUNCTION


SUB UTF8TOUTF16(BYVAL source AS STRING, BYREF destination AS WSTRING, BYVAL size AS INTEGER)
  DIM AS INTEGER i, i2, l
  DIM AS UBYTE c
  ' #todo:/#note: only working on Little Endian systems
  ' big endian code should convert everything internally into Little Endian encoding

  destination[0]=&hFEFF
  l=LEN(source)
  i=0
  i2=1
  WHILE i2<size AND i<l
    c=source[i]
    IF c>127 THEN
      destination[i2]=UTF8GETCODE(MID(source, i+1))
      ' the first bits are telling how long a UTF-8 code is
      ' (maximum 4 bytes)
      IF (c AND 224)=192 THEN
        i+=2
      ELSEIF (c AND 240)=224 THEN
        i+=3
      ELSE
        i+=4
      ENDIF
    ELSE
      destination[i2]=c
      i+=1
    ENDIF
    i2+=1
  WEND
  IF i2>=size THEN i2=size-1
  destination[i2]=0
END SUB


FUNCTION UTF8ISVALID(BYVAL source AS STRING, BYVAL strstart AS INTEGER=1, BYVAL strend AS INTEGER=-1, BYVAL nullbytesinvalid AS BOOLEAN=FALSE) AS BOOLEAN
  DIM AS INTEGER i, l, j
  DIM AS UBYTE c

  IF strstart<2 THEN i=0 ELSE i=strstart-1 ' also compensates a false '0' position
  IF strend<1 THEN l=LEN(source) ELSE l=strend ' -1, 0 => string length is end position
  IF l>LEN(source) THEN RETURN FALSE ' have not to take all nonsense

  WHILE i<l
    c=source[i] ' this 'c' can only be an ASCII or first UTF-8 character
    IF c<128 THEN ' ASCII range
      IF c=0 AND nullbytesinvalid THEN RETURN FALSE
    ELSE ' reserved for UTF-8 sequences
      IF (c AND 192)<192 THEN RETURN FALSE ' invalid first UTF-8 character

      IF c>244 THEN RETURN FALSE ' invalid UTF-8 range

      IF c>=240 THEN ' start of 4-byte sequence
        IF c=244 THEN
          i+=1
          c=source[i]
          IF (c AND 192)<>128 THEN RETURN FALSE ' invalid following character
          IF c>&h8F THEN RETURN FALSE ' outside valid UTF-8 range
          j=2
        ELSE
          j=3
        ENDIF
      ELSEIF c>=224 THEN ' start of 3-byte sequence
        IF c=239 THEN ' check for non-character codes 0xFFFE and 0xFFFF
          i+=1
          c=source[i]
          IF (c AND 192)<>128 THEN RETURN FALSE ' invalid following character
          IF c=&hBF THEN
            i+=1
            c=source[i]
            IF c=&hBE THEN RETURN FALSE ' THE non-character code 0xFFFE => invalid
            IF c=&hBF THEN RETURN FALSE ' 0xFFFF is also an invalid Unicode char
            j=0
          ELSE
            j=1
          ENDIF
        ELSE ' other 3-byte sequences
          j=2
        ENDIF
      ELSE ' => start of 2-byte sequence
        j=1
      ENDIF
      ' j is a repeat count for checking UTF-8 "following"/"secondary" characters
      ' which all have to have the highest two bits being "10"
      ' j=3 is for a 4 byte long UTF-8 sequence, the first byte has its first/highest bits
      ' being "11110", the following bytes the first bits "10"
      WHILE j
        i+=1
        c=source[i]
        IF (c AND 192)<>128 THEN RETURN FALSE ' invalid following character
        j-=1
      WEND
      i+=1
      c=source[i]
    ENDIF
    i+=1
  WEND

  ' nothing invalid found so far
  RETURN TRUE
END FUNCTION


' returning character width of a Unicode character in (Latin) 'characters' / half squares
' i.e. 0(no output), 1, or 2 (double-width, square character)
FUNCTION UNICODECHARWIDTH(BYREF character AS ULONG) AS UBYTE
  IF character<128 THEN ' ASCII range, not recommended to use this function for
    IF character>31 THEN RETURN 1 ' normal chars + Space
    RETURN 0 ' control characters + Nullbyte => "invisible"
  ENDIF

  ' exclude control characters 0x80-0x9F + Soft Hyphen 0xAD from counting
  IF character<160 OR character=&hAD THEN RETURN 0

  ' #todo: determine/exclude all double-width chars as fast as possible

  IF character=&hFEFF THEN RETURN 0 ' BOM code (should be excluded without using UNICODECHARWIDTH())
  RETURN 1 ' default

  ' kind of error, maybe illegal characters, normally shouldn't ever be reached
  RETURN 0
END FUNCTION


' returning character width of a Unicode character in (Latin) 'characters' / half squares
' i.e. 0(no output), 1, or 2 (double-width, square character)
FUNCTION UNICODECHARWIDTH(BYREF source AS STRING, BYREF position AS INTEGER) AS UBYTE
  DIM AS INTEGER i
  DIM AS UBYTE c

  IF position<1 OR source="" THEN RETURN 0 ' invalid
  IF position>LEN(source) THEN RETURN 0 ' error

  i=position-1
  c=source[i]
  IF c<192 THEN ' UTF8 sequence byte, invalid
    ' get next valid character
    i+=1
    IF i>=LEN(source) THEN RETURN 0 ' error
    DO
      c=source[i]
      IF c<128 OR c>191 THEN EXIT DO
      i+=1
      IF i>=LEN(source) THEN RETURN 0 ' error
    LOOP
  ENDIF

  IF c<128 THEN ' ASCII
    IF c>31 THEN
      RETURN 1 ' normal ASCII code
    ELSE
      RETURN 0 ' invisible control code
    ENDIF
  ENDIF

  IF c=&hC2 THEN ' 128  192
    i+=1
    IF i>=LEN(source) THEN RETURN 0 ' error
    c=source[i]
    IF c<160 THEN ' U+0080...U+009F
      RETURN 0 ' formatting characters (sizeless)
    ELSE ' U+00A0...U+00BF (latin letters)
      IF c=&hAD THEN RETURN 0 ' soft hyphen SHY
      RETURN 1
    ENDIF
  ELSEIF c<&hCC THEN ' U+00C0...U+02FF (latin letters, IPA...)
    RETURN 1
  ENDIF

  IF c<&hCD THEN RETURN 0 ' Combining Diacritical Marks
  IF c=&hCD THEN
    i+=1
    IF i>=LEN(source) THEN RETURN 0 ' error
    c=source[i]
    IF c<&hB0 THEN RETURN 0 ' Combining Diacritical Marks
    RETURN 1 ' U+0340...U+037F
  ENDIF

  ' #todo: determine/exclude all double-width chars as fast as possible => "RETURN 2"

  IF c=&hEF THEN
    ' BOM code U+FEFF (should be excluded without using UNICODECHARWIDTH())
    i+=1
    IF i>=LEN(source) THEN RETURN 0 ' error
    c=source[i]
    IF c=&hBB THEN
      i+=1
      IF i>=LEN(source) THEN RETURN 0 ' error
      c=source[i]
      IF c=&hBF THEN RETURN 0 ' BOM is sizeless
    ENDIF
  ENDIF
  RETURN 1 ' default

  ' kind of error, maybe illegal characters, normally shouldn't ever be reached
  RETURN 0
END FUNCTION



FUNCTION UTF8LEN(BYVAL source AS STRING) AS INTEGER
  DIM AS INTEGER i, l, result
  DIM AS UBYTE c

  ' ignore a leading BOM code (BOM makes no sense in UTF-8)
  ' BOM is in UTF-8 always the same, even from big endian platforms
  IF MID(source, 1, 3)=CHR(&hEF, &hBB, &hBF) THEN i=4 ELSE i=0

  result=0
  l=LEN(source)
  WHILE i<l
    c=source[i]
    IF c>127 THEN ' UTF8 encoding
      IF c AND 64 THEN ' => starting byte of UTF8 code = new character
        IF c=&hC3 THEN
          i+=1
          c=source[i]
          ' exclude control characters 0x80-0x9F + soft hyphen 0xAD from counting
          IF (c<>&hAD AND c>&h9F) THEN
            result+=1
          ENDIF
        ELSE
          ' much slower but complete check for the rest:
          result+=UNICODECHARWIDTH(UTF8GETCODE(source))
        ENDIF
      ENDIF
    ELSE ' ASCII "encoding"
      IF c<32 THEN ' ASCII control character
        IF c=10 OR c=13 THEN ' line break codes
          result+=1
          ' #note: all other codes - even tabulator and nullbyte - are completely ignored
          ' a DOS line break will extend the len value by 2 because it's "\r\n"; so the string
          ' should be cleared from this or all line break codes first
        ENDIF
      ELSE
        result+=1
      ENDIF
    ENDIF
    i+=1
  WEND

  RETURN result
END FUNCTION

FUNCTION UTF8MID(BYREF source AS STRING, position AS INTEGER, offset AS INTEGER, length AS INTEGER=-1) AS STRING
  DIM AS INTEGER i, l, l8, j
  DIM AS ULONG uc
  DIM AS UBYTE c
  DIM AS STRING result

  IF source="" OR position<1 OR length<-1 THEN RETURN "" ' error, invalid argument

  IF length<1 THEN ' -1 = "infinite" => return full rest of string (0 works like -1, too)
    length=LEN(source) ' (byte count can't be less than character count)
    ' (not even if nothing but double-width chars, if you won't create such in the ASCII range)
  ENDIF

  result=""
  i=position-1
  l=LEN(source) ' source string length in bytes
  l8=0 ' current substring length

  IF position=LEN(source) THEN
    ' Special case: position is LEN(source). This should address the last character,
    ' but acutally addresses the last byte
    DO
      c=source[i]
      IF c<128 OR c>191 THEN EXIT DO
      i-=1
      ' #note: it's in principle impossible that i becomes less than 0,
      '        requires a corrupt string
    LOOP
  ENDIF

  ' modify (byte) position by +/- characters
  IF offset<>0 THEN
    IF offset<0 THEN ' include leading chars
      WHILE i
        i-=1
        c=source[i]
        IF c<128 OR c>191 THEN ' ASCII or UTF-8 start byte
          ' #note: 192, 193 and 245-255 are illegal values
          offset+=1
          IF offset=0 THEN EXIT WHILE
        ENDIF
      WEND
    ELSE ' start with following chars
      i+=1
      WHILE i<l
        c=source[i]
        IF c<128 OR c>191 THEN ' ASCII or UTF-8 start byte
          ' #note: 192, 193 and 245-255 are illegal values
          offset-=1
          IF offset=0 THEN EXIT WHILE
        ENDIF
        i+=1
      WEND
    ENDIF
    IF offset<>0 THEN RETURN "" ' error
  ENDIF

  IF i=l THEN RETURN "" ' error

  WHILE i<l
    c=source[i]
    IF c<128 THEN ' ASCII range
      l8+=1
      result+=CHR(c)
      i+=1
    ELSE ' UTF-8 range or invalid
      IF c<192 THEN ' inmidst a code sequence, proceed to next valid char (or end)
        i+=1
        WHILE i<l
          c=source[i]
          i+=1
          IF c>191 OR c<128 THEN EXIT WHILE ' start byte or ASCII
        WEND
      ELSE ' => UTF-8 startbyte
        ' #note: the values 192, 193 and 246-255 would be illegal,
        '        expecting valid UTF-8
        ' #todo: maybe catch all invalid codes
        i+=1 ' (to MID() position)
      ENDIF
      IF i>=l THEN RETURN result ' last byte reached (data error)
      ' note: source[i] is the byte before the MID()-index, i has to be 1 bigger here than with "source[i]"
      uc=UTF8GETCODE(source, i)
      j=UNICODECHARWIDTH(uc)
      IF j=2 THEN ' double width char
        l8+=2
        IF l8>length THEN RETURN result ' length exceeded, abort
        result+=UTF8(uc)
      ELSE ' single width or invisible char
        l8+=j
        result+=UTF8(uc)
      ENDIF
      ' skip sequence bytes
      WHILE i<l
        c=source[i]
        IF c<128 OR c>191 THEN EXIT WHILE
        i+=1
      WEND
    ENDIF
    IF l8>=length THEN RETURN result
  WEND

  RETURN result
END FUNCTION


FUNCTION UTF8LEFT(BYVAL source AS STRING, toposition AS INTEGER, offset AS INTEGER=0) AS STRING
  DIM AS INTEGER i, l, l8, j
  DIM AS ULONG uc
  DIM AS UBYTE c
  DIM AS STRING result

  IF source="" OR toposition<1 OR (toposition+offset<1) THEN RETURN "" ' illegal or nonsense values

  ' test next and following chars - get end of UTF-8 sequence
  ' #note: does not really test for valid UTF-8 codes, just start of a new character
  '        => only works with a valid UTF-8 string
  l=LEN(source)
  i=toposition
  WHILE i<l
    c=source[i]
    IF c<128 OR c>191 THEN EXIT WHILE ' ASCII or start byte
    i+=1
  WEND
  ' (#note: at this point, i is correct MID() position of last byte, not byte index)
  IF offset>-1 THEN ' no offset or append more characters
    IF offset=0 THEN
      ' so far that simple (offset=0 is simply like LEFT(), just not UTF-8 safe)
      result=LEFT(source, i)
      RETURN result
    ENDIF
    WHILE i<l AND offset<>0
      c=source[i]
      i+=1
      IF c<128 OR c>191 THEN
        uc=UTF8GETCODE(source, i)
        j=UNICODECHARWIDTH(uc)
        IF j=2 AND offset=1 THEN EXIT WHILE ' next char too broad
        offset-=1
      ENDIF
    WEND
    result=LEFT(source, i)
  ELSE
    WHILE i
      i-=1
      c=source[i]
      IF c<128 OR c>191 THEN ' ASCII or start byte
        uc=UTF8GETCODE(source, i)
        j=UNICODECHARWIDTH(uc)
        IF j=2 AND offset=-1 THEN EXIT WHILE ' next char too broad
        offset+=j
      ENDIF
      IF offset>=0 THEN EXIT WHILE
    WEND
    result=LEFT(source, i)
  ENDIF

  RETURN result
END FUNCTION


FUNCTION UTF8RIGHT(BYVAL source AS STRING, length AS INTEGER) AS STRING
  DIM AS INTEGER i
  DIM AS STRING result
  DIM AS UBYTE c

  IF length=0 OR source="" THEN RETURN "" ' easy task

  ' checking backwards and counting full characters until
  ' length in characters is reached
  i=LEN(source)
  DO
    i-=1
    c=source[i]
    IF c<128 THEN ' ASCII
      length-=1
    ELSEIF c>191 THEN ' UTF-8 startbyte
      IF UNICODECHARWIDTH(UTF8GETCODE(source,i+1))=2 THEN
        ' don't return too much chars => discard exceeding double-width chars
        length-=2
        IF length<0 THEN
          ' ensure, that "i" points to a valid char (start) rather than any mid-byte
          i+=1
          WHILE i<LEN(source)
            c=source[i]
            IF c<128 OR c>191 THEN EXIT WHILE
            i+=1
          WEND
          IF i=LEN(source) THEN RETURN "" ' kinda error
          EXIT DO
        ENDIF
      ELSE
        length-=1
      ENDIF
    ENDIF
    IF length=0 THEN EXIT DO
  LOOP WHILE i ' (begin of source has been reached)
  ' gets the result string
  IF i THEN ' found a position >0
    result=MID(source, i+1) ' i is current mem.offset, not MID()-position(=> +1)
    ' #todo: ? maybe code manually, but would only make sense in Assembler
    '        => direct memcopy, up to 64 bit, not a "result=result+chr(c)"...
  ELSE
    result=source
  ENDIF

  RETURN result
END FUNCTION


FUNCTION UTF8DIVIDETOLINES(BYREF text AS STRING, BYVAL linewidth AS INTEGER) AS STRING
  DIM AS INTEGER i, l, l2, minl, oldi, oldl2
  DIM AS ULONG uc

  IF linewidth<6 OR linewidth>160 THEN
    ' illegal(really stupid) parameter value, error
    text=""
    RETURN text
  ENDIF

  ' minimum line width
  ' - only for auto-breaking
  '   minl is used to determine whether to break a superlong
  '   word or to put it into the next line, to not let it
  '   look awkward
  '   an automatically broken line must at least have
  '   50% of a full line. Superlong words may have to
  '   be broken forcedly
  '   the shorter the lines, the broader the "white space"
  '   can be percentually
  IF linewidth>32 THEN
    IF linewidth<80 THEN
      minl=linewidth-16 ' last word in a line may be 16 broad
    ELSE
      minl=(linewidth*4)/5 ' 80%
    ENDIF
  ELSE
    minl=(linewidth+1) SHR 1 ' 50% (i.e. 3-16 chars)
  ENDIF

  ' processing string, adding code 13 for new lines
  i=0
  l=LEN(text)
  l2=0
  WHILE i<l
    uc=text[i] ' 1 byte, not Unicode
    i+=1
    IF uc<128 THEN ' ASCII range
      IF uc>31 THEN ' normal character, width 1
        l2+=1
      ELSEIF uc=13 THEN ' line break => new line (forced)
        l2=0
        ' ELSE: control character with no width (l2+0)
      ENDIF
    ELSE
      IF (uc AND 192)=192 THEN
        uc=UTF8GETCODE(text, i)
        IF uc<>&hAD AND uc>=160 THEN
          ' #todo: faster char width pre-check
          l2+=UNICODECHARWIDTH(uc)
        ENDIF
      ENDIF
    ENDIF

    IF l2>linewidth THEN
      ' linewidth reached
      ' note: if last char has double width, l2 can be linewidth+2 instead of +1
      oldi=i
      oldl2=l2
      WHILE l2>minl
        i-=1
        uc=text[i] ' 1 byte, not Unicode
        IF uc<128 THEN ' ASCII range
          IF uc=32 THEN ' Space
            l2-=2
            IF l2<=linewidth THEN
              ' make a new line
              text[i]=13
              l2+=2
              i+=1
              EXIT WHILE
            ENDIF
            l2+=1
          ELSEIF uc=45 THEN ' hyphen
            l2-=1
            IF l2<=linewidth THEN
              i+=1
              ' insert line break after hyphen
              text=LEFT(text, i)+CHR(13)+MID(text, i+1)
              l+=1
              i+=1
              EXIT WHILE
            ENDIF
            l2+=1
          ENDIF
          IF uc>32 THEN l2-=1
        ELSEIF (uc AND 192)=192 THEN ' UTF-8 sequence start
          uc=UTF8GETCODE(text, i+1)
          IF uc=&hAD THEN ' Soft Hyphen SHY, invisible character
            IF l2<linewidth THEN
              ' replace (2 bytes) with hyphen+line break
              text[i]=45
              i+=1
              text[i]=13
              EXIT WHILE
            ENDIF
            ' (l2-=0)
          ELSE
            ' #todo: faster char width pre-check
            l2-=UNICODECHARWIDTH(uc)
          ENDIF
        ENDIF
      WEND
      IF l2<=minl THEN ' last word is too long
        i=oldi
        l2=oldl2
        WHILE i AND l2>=linewidth
          i-=1
          uc=text[i]
          IF uc<128 THEN
            l2-=1 ' (can't be a control character, not even Space)
          ELSEIF (uc AND 192)=192 THEN
            uc=UTF8GETCODE(text, i+1)
            ' #todo: faster char width pre-check
            l2-=UNICODECHARWIDTH(uc)
          ENDIF
        WEND
        ' maximum word length, merge hyphen and line break
        uc=text[i]
        IF uc>127 THEN ' UTF-8 char, go to end of sequence
          DO
            i+=1
            uc=text[i]
          LOOP WHILE (uc AND 192)=128
        ENDIF
        text=LEFT(text, i)+CHR(45, 13)+MID(text, i+1)
      ENDIF
      l2=0 ' new line in all cases
    ENDIF
  WEND

  RETURN text
END FUNCTION


FUNCTION UNICODEUPPERCHAR(unicode AS ULONG) AS ULONG
  IF unicode<128 THEN ' ASCII range
    IF unicode>96 AND unicode<123 THEN ' ASCII a-z
      RETURN unicode AND 95 ' clear bit 5 (&7)
    ENDIF
    ' all others haven't to be changed
    RETURN unicode
  ELSE
    IF unicode<&h100 THEN ' Latin 1 range (or lower)
      IF unicode>&hDE THEN ' minuscules
        ' except "sharp S", the division sign and "Y+diaeresis"
        IF unicode>&hDF AND unicode<>&hF7 AND unicode<&hFF THEN
          unicode AND=223 ' clear bit 5 (of 8 bits)
          RETURN unicode
        ELSE
          ' "sharp S" is a minuscule, the division sign... a sign
          IF unicode=&hFF THEN
            ' big Y+diaeresis (no word begins with a Y+diaeresis)
            RETURN &h178
          ELSEIF unicode=&hDF THEN
            ' newly introduced big "Sharp S" - normally mostly double-S
            ' would be used (but not correctly translated in minuscule)
            ' => might require extra treatment of the programmer in case
            RETURN &h1E9E
          ENDIF
        ENDIF
      ENDIF
    ELSEIF unicode<&h250 THEN ' Latin Extended A+B
      IF unicode<&h138 THEN
        IF (unicode AND 1) THEN
          unicode-=1 ' minuscule -> majuscule
          RETURN unicode
        ENDIF
      ELSEIF unicode=&h138 THEN
        ' "Kra" is a minuscule without majuscule, normally translated as K'
        ' (K with apostrophe);
        ' => won't be retranslated into "Kra" minuscule
        RETURN &h39A ' Code for Kappa (Greek), => also using 2 bytes
      ELSEIF unicode<&h149 THEN
        IF (unicode AND 1)=0 THEN
          unicode-=1 ' minuscule -> majuscule
          RETURN unicode
        ENDIF
      ELSEIF unicode=&h149 THEN
        ' not even a letter that should exist in Unicode and shouldn't be used
        ' following the guidelines => ignore
      ELSEIF unicode<&h178 THEN
        IF (unicode AND 1) THEN
          unicode-=1 ' minuscule -> majuscule
          RETURN unicode
        ENDIF
      ELSEIF unicode=&h178 THEN
        ' Y+diaeresis majuscule => ignore
      ELSEIF unicode<&h17F THEN
        IF (unicode AND 1)=0 THEN
          unicode-=1 ' minuscule -> majuscule
          RETURN unicode
        ENDIF
      ELSEIF unicode=&h17F THEN
        ' long small letter S "small inner S"; has to be translated to big S, but
        ' won't be retranslated correctly, instead to a normal ("ending") "s"
        RETURN 83 ' ASCII Code for S
      ELSEIF unicode<&h1DE THEN
        ' official letter madness block it seems...
        SELECT case AS CONST unicode
          CASE &h183, &h185, &h188, &h18C, &h18F, &h192, &h199, &h1A1, &h1A3, _
            &h1A5, &h1A8, &h1B0, &h1B4, &h1B6, &h1B9, &h1BD, &h1C5, &h1C8, &h1CB
            RETURN unicode-1
          CASE &h1C6, &h1C9, &h1CC
            RETURN unicode-2
          CASE &h1CE TO &h1DC
            IF (unicode AND 1)=0 THEN
              RETURN unicode-1
            ENDIF
        END SELECT
      ELSEIF unicode<&h1F0 THEN
        IF (unicode AND 1) THEN
          RETURN unicode-1
        ENDIF
      ELSEIF unicode=&h1F0 THEN
        ' ignore
      ELSEIF unicode<&h1F8 THEN
        IF unicode=&h1F2 OR unicode=&h1F3 THEN
          RETURN &h1F1
        ELSEIF unicode=&h1F5 THEN
          RETURN &h1F4
        ENDIF
      ELSEIF unicode<&h234 THEN
        IF unicode<&h220 OR unicode>&h221 THEN
          IF (unicode AND 1) THEN
            RETURN unicode-1
          ENDIF
        ENDIF
      ELSEIF unicode<&h246 THEN
        IF unicode=&h23C OR unicode=&h242 THEN
          RETURN unicode-1
        ENDIF
      ELSE
        IF (unicode AND 1) THEN
          RETURN unicode-1
        ENDIF
      ENDIF
    ELSEIF unicode<&h370 THEN ' IPA chars and special chars with no upper/lower case
      RETURN unicode ' => ignore
    ELSEIF unicode<&h400 THEN ' Greek and Coptic letters
      IF unicode<&h386 THEN
        IF unicode=&h371 OR unicode=&h373 OR unicode=&h377 THEN
          RETURN unicode-1
        ELSEIF unicode>=&h37B AND unicode<=&h37D THEN
          RETURN unicode+130
        ENDIF
      ELSEIF unicode<&h3AC THEN
        ' ignore
      ELSEIF unicode<&h3B0 THEN
        IF unicode=&h3AC THEN
          RETURN &h386
        ELSE
          RETURN unicode-37
        ENDIF
      ELSEIF unicode<&h3CC THEN
        IF unicode>&h3B0 THEN
          unicode AND=&h3DF ' minuscule -> majuscule
          IF unicode=&h3A2 THEN unicode=&h3A3 ' => Sigma
          RETURN unicode
        ENDIF
      ELSEIF unicode<&h3D8 THEN
        IF unicode=&h3CC THEN
          RETURN &h38C
        ELSEIF unicode=&h3CD OR unicode=&h3CE THEN
          RETURN unicode-31 ' minuscule -> majuscule
        ELSEIF unicode=&h3D7 THEN
          RETURN &h3CF
        ENDIF
      ELSEIF unicode<&h3F0 THEN
        IF (unicode AND 1) THEN
          RETURN unicode-1
        ENDIF
      ELSE
        IF unicode=&h3F2 THEN ' Lunate Sigma
          RETURN &h3F9
        ELSEIF unicode=&h3F3 THEN ' Jot
          RETURN &h37F
        ELSEIF unicode=&h3F8 OR unicode=&h3FB THEN
          RETURN unicode-1
        ENDIF
      ENDIF
    ELSEIF unicode<&h530 THEN ' Cyrillic letters
      IF unicode<&h430 THEN
        ' majuscules, ignore
      ELSEIF unicode<&h450 THEN
        unicode AND=&h7DF ' minuscule -> majuscule
        RETURN unicode
      ELSEIF unicode<&h460 THEN
        unicode AND=&h7AF ' minuscule -> majuscule
        RETURN unicode
      ELSEIF unicode<&h482 THEN
        IF (unicode AND 1) then
          RETURN unicode-1 ' minuscule -> majuscule
        ENDIF
      ELSEIF unicode<&h48A THEN
        ' ignore
      ELSEIF unicode<&h4C0 THEN
        IF (unicode AND 1) then
          RETURN unicode-1 ' minuscule -> majuscule
        ENDIF
      ELSEIF unicode=&h4C0 THEN
        ' ignore
      ELSEIF unicode<&h4CF THEN
        IF (unicode AND 1)=0 then
          RETURN unicode-1 ' minuscule -> majuscule
        ENDIF
      ELSEIF unicode>&h4CF THEN
        IF (unicode AND 1) then
          RETURN unicode-1 ' minuscule -> majuscule
        ENDIF
      ENDIF
    ELSE
      ' => no upper/lower case - maybe, / not yet supported
    ENDIF
  ENDIF

  RETURN unicode ' default / ignored codes
END FUNCTION


FUNCTION UNICODELOWERCHAR(unicode AS ULONG) AS ULONG
  IF unicode<128 THEN ' ASCII range
    IF unicode>64 AND unicode<91 THEN ' ASCII A-Z
      RETURN unicode OR 32 ' => a-z
    ENDIF
    ' all others haven't to be changed
    RETURN unicode
  ELSE
    IF unicode<&h100 THEN ' Latin 1 range (or lower)
      IF unicode>=&hC0 AND unicode<=&hDE THEN ' majuscules
        ' except multiplication sign
        IF unicode<>&hD7 THEN
          unicode OR=32 ' turn to minuscules
          RETURN unicode
        ENDIF
      ENDIF
    ELSEIF unicode<&h250 THEN ' Latin Extended A+B
      IF unicode<&h138 THEN
        IF (unicode AND 1)=0 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h138 THEN
        ' ("Kra" is a minuscule without majuscule), ignore
      ELSEIF unicode<&h149 THEN
        IF (unicode AND 1) THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h149 THEN
        ' not even a letter that should exist in Unicode and shouldn't be used
        ' following the guidelines
        ' => ignore
      ELSEIF unicode<&h178 THEN
        IF (unicode AND 1)=0 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h178 THEN
        ' Y+diaeresis majuscule
        RETURN &hFF
      ELSEIF unicode<&h17F THEN
        IF (unicode AND 1) THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h17F THEN
        ' long small letter S => ignore
      ELSEIF unicode<&h1DE THEN
        ' official letter madness block it seems...
        SELECT case AS CONST unicode
          CASE &h182, &h184, &h187, &h18B, &h18E, &h191, &h198, &h1A0, &h1A2, _
            &h1A4, &h1A7, &h1AF, &h1B3, &h1B5, &h1B8, &h1BC, &h1C5, &h1C8, &h1CB
            RETURN unicode+1 ' majuscule -> minuscule
          CASE &h1C4, &h1C7, &h1CA
            RETURN unicode+2 ' majuscule -> minuscule
          CASE &h1CD TO &h1DB
            IF (unicode AND 1) THEN
              RETURN unicode+1 ' majuscule -> minuscule
            ENDIF
        END SELECT
      ELSEIF unicode<&h1F0 THEN
        IF (unicode AND 1)=0 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h1F0 THEN
        ' ignore
      ELSEIF unicode<&h1F8 THEN
        IF unicode=&h1F1 OR unicode=&h1F2 THEN
          RETURN &h1F3
        ELSEIF unicode=&h1F4 THEN
          RETURN &h1F5
        ENDIF
      ELSEIF unicode<&h234 THEN
        IF unicode<&h220 OR unicode>&h221 THEN
          IF (unicode AND 1)=0 THEN
            RETURN unicode+1 ' majuscule -> minuscule
          ENDIF
        ENDIF
      ELSEIF unicode<&h246 THEN
        IF unicode=&h23B OR unicode=&h241 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSE
        IF (unicode AND 1)=0 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ENDIF
    ELSEIF unicode<&h370 THEN ' IPA chars and special chars with no upper/lower case
      ' ignore
    ELSEIF unicode<&h400 THEN ' Greek and Coptic letters
      IF unicode<&h386 THEN
        IF unicode=&h370 OR unicode=&h372 OR unicode=&h376 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ELSEIF unicode=&h37F THEN ' Jot
          RETURN &h3F3
        ENDIF
      ELSEIF unicode<&h391 THEN
        IF unicode=&h386 THEN
          RETURN &h3AC
        ELSEIF unicode>&h387 AND unicode<&h38B THEN
          RETURN unicode+37 ' majuscule -> minuscule
        ELSEIF unicode=&h38C THEN
          RETURN &h3CC
        ELSEIF unicode=&h38E OR unicode=&h38F THEN
          RETURN unicode+63 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode<&h3AC THEN
        ' core alphabet
        RETURN unicode OR 32 ' majuscule -> minuscule
        ' #note: there are two versions of Sigma: &h3C3 and &h3C2
        '        last one is for word endings, can't determine this here
      ELSEIF unicode<&h3CF THEN
        ' ignore
      ELSEIF unicode<&h3D8 THEN
        IF unicode=&h3CF THEN
          RETURN &h3D7
        ENDIF
      ELSEIF unicode<&h3F0 THEN
        IF (unicode AND 1)=0 THEN
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSE
        IF unicode=&h3F7 THEN
          RETURN &h3F8
        ELSEIF unicode=&h3F9 THEN ' Lunate Sigma
          RETURN &h3F2
        ELSEIF unicode=&h3FA THEN
          RETURN &h3FB
        ELSEIF unicode>=&h3FD AND unicode<=&h3FF THEN ' Lunate Sigma
          RETURN unicode+130 ' majuscule -> minuscule
        ENDIF
      ENDIF
    ELSEIF unicode<&h530 THEN ' Cyrillic letters
      IF unicode<&h410 THEN
        RETURN unicode OR &h50 ' minuscule -> majuscule
      ELSEIF unicode<&h430 THEN
        RETURN unicode OR 32 ' minuscule -> majuscule
      ELSEIF unicode<&h460 THEN
        ' minuscules => ignore
      ELSEIF unicode<&h482 THEN
        IF (unicode AND 1)=0 then
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode<&h48A THEN
        ' ignore
      ELSEIF unicode<&h4C0 THEN
        IF (unicode AND 1)=0 then
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode=&h4C0 THEN
        ' ignore
      ELSEIF unicode<&h4CF THEN
        IF (unicode AND 1) then
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ELSEIF unicode>&h4CF THEN
        IF (unicode AND 1)=0 then
          RETURN unicode+1 ' majuscule -> minuscule
        ENDIF
      ENDIF
    ELSEIF unicode=&h1E9E THEN ' big "sharp S"
      RETURN &hDF
    ELSE
      ' => no upper/lower case - maybe, / not yet supported
    ENDIF
  ENDIF

  RETURN unicode ' default / ignored codes
END FUNCTION


FUNCTION UTF8UPPERCASE(BYREF ustring AS STRING, BYVAL start AS INTEGER=-1, BYVAL length AS INTEGER=-1) AS STRING
  DIM AS INTEGER i, l
  DIM AS UBYTE c
  DIM AS ULONG uc, oc
  DIM AS STRING code

  ' "smart" parameters: U.C.(string, start, length) <-> U.C.(string, length) instead of U.C.(string, start)
  IF start=-1 AND length=-1 THEN
    i=0
    l=LEN(ustring)
  ELSE
    IF length=-1 THEN
      i=0
      l=start
    ELSE
      IF start>0 THEN i=start-1 ELSE i=0
      IF length>0 THEN l=length ELSE l=LEN(ustring)
    ENDIF
  ENDIF

  WHILE i<l
    c=ustring[i]
    IF c<128 THEN ' ASCII range
      IF (c>96 AND c<123) THEN ustring[i]=(c AND 95) ' a-z => A-Z
      i+=1
    ELSE ' UTF-8 range
      ' #note: invalid UTF-8 will require to remove illegal characters
      ' get Unicode value
      i+=1
      oc=UTF8GETCODE(ustring, i)
      uc=UNICODEUPPERCHAR(oc)
      ' check whether old and new UTF-8 code have the same length
      IF uc>127 then
        IF oc<&h0800 AND uc<&h0800 THEN ' 2 bytes
          MID(ustring,i,2)=UTF8(uc)
          i+=1
        ELSEIF oc>&h07FF AND oc<&h10000 AND uc>&h07FF AND uc<&h10000 THEN ' 3 bytes
          MID(ustring,i,3)=UTF8(uc)
          i+=2
        ELSEIF oc>&hFFFF AND uc>&hFFFF THEN ' 4 bytes
          MID(ustring,i,4)=UTF8(uc)
          i+=3
        ELSE ' UTF-8 codes have different length
          ' universal but slow algorithm
          code=UTF8(uc)
          ustring=LEFT(ustring, i-1)+code+UTF8MID(ustring,i,1)
          i+=len(code)-1
          l=LEN(ustring)
        ENDIF
      ELSE ' uc is 1 byte / ASCII
        ' universal but slow algorithm
        code=UTF8(uc)
        ustring=LEFT(ustring, i-1)+code+UTF8MID(ustring,i,1)
        i+=len(code)-1
        l=LEN(ustring)
      ENDIF
    ENDIF
  WEND

  RETURN ustring
END FUNCTION


FUNCTION UTF8LOWERCASE(BYREF ustring AS STRING, BYVAL start AS INTEGER=-1, BYVAL length AS INTEGER=-1) AS STRING
  DIM AS INTEGER i, l
  DIM AS UBYTE c
  DIM AS ULONG uc, oc
  DIM AS STRING code

  ' "smart" parameters: L.C.(string, start, length) <-> L.C.(string, length) instead of L.C.(string, start)
  IF start=-1 AND length=-1 THEN ' whole string
    i=0
    l=LEN(ustring)
  ELSE
    IF length=-1 THEN ' parameter is length instead of start
      i=0
      l=start
    ELSE
      IF start>0 THEN i=start-1 ELSE i=0
      IF length>0 THEN l=length ELSE l=LEN(ustring)
    ENDIF
  ENDIF

  WHILE i<l
    c=ustring[i]
    IF c<128 THEN ' ASCII range
      IF (c>64 AND c<91) THEN ustring[i]=(c OR 32) ' A-Z => a-z
      i+=1
    ELSE ' UTF-8 range
      ' #note: invalid UTF-8 will require to remove illegal characters
      ' get Unicode value
      i+=1
      oc=UTF8GETCODE(ustring, i)
      uc=UNICODELOWERCHAR(oc)
      ' check whether old and new UTF-8 code have the same length
      IF uc>127 then
        IF oc<&h0800 AND uc<&h0800 THEN ' 2 bytes
          MID(ustring,i,2)=UTF8(uc)
          i+=1
        ELSEIF oc>&h07FF AND oc<&h10000 AND uc>&h07FF AND uc<&h10000 THEN ' 3 bytes
          MID(ustring,i,3)=UTF8(uc)
          i+=2
        ELSEIF oc>&hFFFF AND uc>&hFFFF THEN ' 4 bytes
          MID(ustring,i,4)=UTF8(uc)
          i+=3
        ELSE ' UTF-8 codes have different length
          ' universal but slow algorithm
          code=UTF8(uc)
          ustring=LEFT(ustring, i-1)+code+UTF8MID(ustring,i,1)
          i+=len(code)-1
          l=LEN(ustring)
        ENDIF
      ELSE ' uc is 1 byte / ASCII
        ' universal but slow algorithm
        code=UTF8(uc)
        ustring=LEFT(ustring, i-1)+code+UTF8MID(ustring,i,1)
        i+=len(code)-1
        l=LEN(ustring)
      ENDIF
    ENDIF
  WEND

  RETURN ustring
END FUNCTION
